@@ -5,17 +5,19 @@ use anyhow::Result;
 use clap::Parser;
 use pipewire as pw;
 use spa::pod::Pod;
+use std::rc::Rc;
 use std::{cell::RefCell, collections::HashMap};
-use std::{rc::Rc, sync::Arc};
 
-use pw::link::Link;
-use pw::metadata::Metadata;
-use pw::node::Node;
-use pw::port::Port;
-use pw::properties;
-use pw::proxy::{Listener, ProxyListener, ProxyT};
-use pw::types::ObjectType;
-use pw::Signal;
+use pw::{
+    link::Link,
+    loop_::Signal,
+    metadata::Metadata,
+    node::Node,
+    port::Port,
+    properties::properties,
+    proxy::{Listener, ProxyListener, ProxyT},
+    types::ObjectType,
+};
 
 struct Proxies {
     proxies_t: HashMap<u32, Box<dyn ProxyT>>,
@@ -54,22 +56,24 @@ impl Proxies {
 }
 
 fn monitor(remote: Option<String>) -> Result<()> {
-    let main_loop = pw::MainLoop::new()?;
+    let main_loop = pw::main_loop::MainLoop::new(None)?;
 
     let main_loop_weak = main_loop.downgrade();
-    let _sig_int = main_loop.add_signal_local(Signal::SIGINT, move || {
+    let _sig_int = main_loop.loop_().add_signal_local(Signal::SIGINT, move || {
         if let Some(main_loop) = main_loop_weak.upgrade() {
             main_loop.quit();
         }
     });
     let main_loop_weak = main_loop.downgrade();
-    let _sig_term = main_loop.add_signal_local(Signal::SIGTERM, move || {
-        if let Some(main_loop) = main_loop_weak.upgrade() {
-            main_loop.quit();
-        }
-    });
+    let _sig_term = main_loop
+        .loop_()
+        .add_signal_local(Signal::SIGTERM, move || {
+            if let Some(main_loop) = main_loop_weak.upgrade() {
+                main_loop.quit();
+            }
+        });
 
-    let context = pw::Context::new(&main_loop)?;
+    let context = pw::context::Context::new(&main_loop)?;
     let props = remote.map(|remote| {
         properties! {
             *pw::keys::REMOTE_NAME => remote
@@ -97,8 +101,8 @@ fn monitor(remote: Option<String>) -> Result<()> {
         })
         .register();
 
-    let registry = Arc::new(core.get_registry()?);
-    let registry_weak = Arc::downgrade(&registry);
+    let registry = Rc::new(core.get_registry()?);
+    let registry_weak = Rc::downgrade(&registry);
 
     // Proxies and their listeners need to stay alive so store them here
     let proxies = Rc::new(RefCell::new(Proxies::new()));
