@@ -1,22 +1,22 @@
-use super::stream::Stream;
+use super::stream::StreamRef;
 
-use spa::data::Data;
+use spa::buffer::Data;
 use std::convert::TryFrom;
 use std::ptr::NonNull;
 
-pub struct Buffer<'s, D> {
+pub struct Buffer<'s> {
     buf: NonNull<pw_sys::pw_buffer>,
 
     /// In Pipewire, buffers are owned by the stream that generated them.
     /// This reference ensures that this rule is respected.
-    stream: &'s Stream<D>,
+    stream: &'s StreamRef,
 }
 
-impl<D> Buffer<'_, D> {
+impl Buffer<'_> {
     pub(crate) unsafe fn from_raw(
         buf: *mut pw_sys::pw_buffer,
-        stream: &Stream<D>,
-    ) -> Option<Buffer<'_, D>> {
+        stream: &StreamRef,
+    ) -> Option<Buffer<'_>> {
         NonNull::new(buf).map(|buf| Buffer { buf, stream })
     }
 
@@ -36,9 +36,14 @@ impl<D> Buffer<'_, D> {
 
         slice_of_data
     }
+
+    #[cfg(feature = "v0_3_49")]
+    pub fn requested(&self) -> u64 {
+        unsafe { self.buf.as_ref().requested }
+    }
 }
 
-impl<D> Drop for Buffer<'_, D> {
+impl Drop for Buffer<'_> {
     fn drop(&mut self) {
         unsafe {
             self.stream.queue_raw_buffer(self.buf.as_ptr());
