@@ -13,7 +13,7 @@
 // ignored because https://gitlab.freedesktop.org/pipewire/pipewire-rs/-/issues/19
 //! ```no_run
 //! use std::{time::Duration, sync::mpsc, thread};
-//! use pipewire::MainLoop;
+//! use pipewire::main_loop::MainLoop;
 //!
 //! // Our message to the pipewire loop, this tells it to terminate.
 //! struct Terminate;
@@ -42,16 +42,16 @@
 //!     main_sender: mpsc::Sender<String>,
 //!     pw_receiver: pipewire::channel::Receiver<Terminate>
 //! ) {
-//!     let mainloop = MainLoop::new().expect("Failed to create main loop");
+//!     let mainloop = MainLoop::new(None).expect("Failed to create main loop");
 //!
 //!     // When we receive a `Terminate` message, quit the main loop.
-//!     let _receiver = pw_receiver.attach(&mainloop, {
+//!     let _receiver = pw_receiver.attach(mainloop.loop_(), {
 //!         let mainloop = mainloop.clone();
 //!         move |_| mainloop.quit()
 //!     });
 //!
 //!     // Every 100ms, send `"Hello"` to the main thread.
-//!     let timer = mainloop.add_timer(move |_| {
+//!     let timer = mainloop.loop_().add_timer(move |_| {
 //!         main_sender.send(String::from("Hello"));
 //!     });
 //!     timer.update_timer(
@@ -65,13 +65,12 @@
 
 use std::{
     collections::VecDeque,
-    ffi::c_void,
     os::unix::prelude::*,
     sync::{Arc, Mutex},
 };
 
-use crate::{IoSource, LoopRef};
-use spa::flags::IoFlags;
+use crate::loop_::{IoSource, LoopRef};
+use spa::support::system::IoFlags;
 
 /// A receiver that has not been attached to a loop.
 ///
@@ -90,22 +89,15 @@ impl<T: 'static> Receiver<T> {
         F: Fn(T) + 'static,
     {
         let channel = self.channel.clone();
-        let eventfd = channel.lock().expect("Channel mutex lock poisoned").eventfd;
+        let readfd = channel.lock().expect("Channel mutex lock poisoned").readfd;
 
-        // Attach the eventfd as an IO source to the loop.
-        // Whenever the eventfd is signaled, call the users callback with each message in the queue.
-        let iosource = loop_.add_io(eventfd, IoFlags::IN, move |_| {
+        // Attach the pipe as an IO source to the loop.
+        // Whenever the pipe is written to, call the users callback with each message in the queue.
+        let iosource = loop_.add_io(readfd, IoFlags::IN, move |_| {
             let mut channel = channel.lock().expect("Channel mutex lock poisoned");
 
-            // Read from the eventfd to make it block until written to again.
-            unsafe {
-                let mut _eventnum: u64 = 0;
-                libc::read(
-                    channel.eventfd,
-                    &mut _eventnum as *mut u64 as *mut c_void,
-                    std::mem::size_of::<u64>(),
-                );
-            }
+            // Read from the pipe to make it block until written to again.
+            let _ = nix::unistd::read(channel.readfd, &mut [0]);
 
             channel.queue.drain(..).for_each(&callback);
         });
@@ -119,7 +111,7 @@ impl<T: 'static> Receiver<T> {
 
 /// A [`Receiver`] that has been attached to a loop.
 ///
-/// Dropping this will cause it to be deattached from the loop, so no more messages will be received.
+/// Dropping this will cause it to be detached from the loop, so no more messages will be received.
 pub struct AttachedReceiver<'l, T>
 where
     T: 'static,
@@ -132,7 +124,7 @@ impl<'l, T> AttachedReceiver<'l, T>
 where
     T: 'static,
 {
-    /// Deattach the receiver from the loop.
+    /// Detach the receiver from the loop.
     ///
     /// No more messages will be received until you attach it to a loop again.
     #[must_use]
@@ -170,16 +162,9 @@ impl<T> Sender<T> {
         // If no messages are waiting already, signal the receiver to read some.
         // Because the channel mutex is locked, it is alright to do this before pushing the message.
         if channel.queue.is_empty() {
-            let res = unsafe {
-                libc::write(
-                    channel.eventfd,
-                    &1u64 as *const u64 as *const c_void,
-                    std::mem::size_of::<u64>(),
-                )
-            };
-            if res == -1 {
-                // Eventfd write failed.
-                return Err(t);
+            match nix::unistd::write(channel.writefd, &[1u8]) {
+                Ok(_) => (),
+                Err(_) => return Err(t),
             }
         }
 
@@ -192,26 +177,26 @@ impl<T> Sender<T> {
 
 /// Shared state between the [`Sender`]s and the [`Receiver`].
 struct Channel<T> {
-    /// A raw eventfd used to signal the loop the receiver is attached to that messages are waiting.
-    eventfd: RawFd,
+    /// A pipe used to signal the loop the receiver is attached to that messages are waiting.
+    readfd: RawFd,
+    writefd: RawFd,
     /// Queue of any messages waiting to be received.
     queue: VecDeque<T>,
 }
 
 impl<T> Drop for Channel<T> {
     fn drop(&mut self) {
-        unsafe {
-            // We do not error check here, because the eventfd does not contain any data that might be lost,
-            // and because there is no way to handle an error in a `Drop` implementation anyways.
-            libc::close(self.eventfd);
-        }
+        // We do not error check here, because the pipe does not contain any data that might be lost,
+        // and because there is no way to handle an error in a `Drop` implementation anyways.
+        let _ = nix::unistd::close(self.readfd);
+        let _ = nix::unistd::close(self.writefd);
     }
 }
 
 /// Create a Sender-Receiver pair, where the sender can be used to send messages to the receiver.
 ///
 /// This functions similar to [`std::sync::mpsc`], but with a receiver that can be attached to any
-/// [`Loop`](`crate::Loop`) to have the loop invoke a callback with any new messages.
+/// [`LoopRef`](`crate::loop_::LoopRef`) to have the loop invoke a callback with any new messages.
 ///
 /// This can be used for inter-thread communication without shared state and where [`std::sync::mpsc`] can not be used
 /// because the receiving thread is running the pipewire loop.
@@ -219,25 +204,11 @@ pub fn channel<T>() -> (Sender<T>, Receiver<T>)
 where
     T: 'static,
 {
-    // Manually open an eventfd that we can use to signal the loop thread to check for messages
-    // via an IoSource.
-    #[cfg(target_os = "linux")]
-    let eventfd = unsafe { libc::eventfd(0, libc::EFD_CLOEXEC) };
-    #[cfg(target_os = "freebsd")]
-    let eventfd = {
-        // Added in FreeBSD 13, libc crate doesn't target that yet
-        use std::os::raw::{c_int, c_uint};
-        extern "C" {
-            pub fn eventfd(name: c_uint, flags: c_int) -> c_int;
-        }
-        unsafe { eventfd(0, 0x00100000) }
-    };
-    if eventfd == -1 {
-        panic!("Failed to create eventfd: {}", nix::errno::errno())
-    }
+    let fds = nix::unistd::pipe2(nix::fcntl::OFlag::O_CLOEXEC).unwrap();
 
     let channel: Arc<Mutex<Channel<T>>> = Arc::new(Mutex::new(Channel {
-        eventfd,
+        readfd: fds.0,
+        writefd: fds.1,
         queue: VecDeque::new(),
     }));
 
