@@ -12,7 +12,7 @@ use std::{
 
 use libc::{c_int, c_void};
 pub use nix::sys::signal::Signal;
-use spa::{flags::IoFlags, result::SpaResult, spa_interface_call_method};
+use spa::{spa_interface_call_method, support::system::IoFlags, utils::result::SpaResult};
 
 use crate::{utils::assert_main_thread, Error};
 
@@ -21,7 +21,7 @@ use crate::{utils::assert_main_thread, Error};
 ///
 /// An owned version, [`Loop`], is available,
 /// which lets you create and own a [`pw_loop`](`pw_sys::pw_loop`),
-/// but other objects, such as [`MainLoop`](`crate::MainLoop`), also contain them.
+/// but other objects, such as [`MainLoop`](`crate::main_loop::MainLoop`), also contain them.
 #[repr(transparent)]
 pub struct LoopRef(pw_sys::pw_loop);
 
@@ -30,20 +30,22 @@ impl LoopRef {
         &self.0
     }
 
-    pub fn as_ptr(&self) -> *mut pw_sys::pw_loop {
-        &self.0 as *const _ as *mut _
+    pub fn as_raw_ptr(&self) -> *mut pw_sys::pw_loop {
+        std::ptr::addr_of!(self.0).cast_mut()
     }
 
     /// Get the file descriptor backing this loop.
-    pub fn fd(&self) -> RawFd {
+    pub fn fd(&self) -> BorrowedFd<'_> {
         unsafe {
             let mut iface = self.as_raw().control.as_ref().unwrap().iface;
 
-            spa_interface_call_method!(
+            let raw_fd = spa_interface_call_method!(
                 &mut iface as *mut spa_sys::spa_interface,
                 spa_sys::spa_loop_control_methods,
                 get_fd,
-            )
+            );
+
+            BorrowedFd::borrow_raw(raw_fd)
         }
     }
 
@@ -129,7 +131,7 @@ impl LoopRef {
     /// Register some type of IO object with a callback that is called when reading/writing on the IO object
     /// is available.
     ///
-    /// The specified `event_mask` determines whether to trigger when either input, output, or any of the two is avaiable.
+    /// The specified `event_mask` determines whether to trigger when either input, output, or any of the two is available.
     ///
     /// The returned IoSource needs to take ownership of the IO object, but will provide a reference to the callback when called.
     #[must_use]
@@ -369,6 +371,14 @@ impl LoopRef {
     }
 }
 
+/// Trait implemented by objects that implement a `pw_loop` and are reference counted in some way.
+///
+/// # Safety
+///
+/// The `LoopRef` returned by the implementation of `AsRef<LoopRef>` must remain valid as long as any clone
+/// of the trait implementor is still alive. \
+pub unsafe trait IsLoopRc: Clone + AsRef<LoopRef> + 'static {}
+
 #[derive(Clone, Debug)]
 pub struct Loop {
     inner: Rc<LoopInner>,
@@ -376,30 +386,54 @@ pub struct Loop {
 
 impl Loop {
     /// Create a new [`Loop`].
-    pub fn new() -> Result<Self, Error> {
-        LoopInner::new().map(|inner| Self {
-            inner: Rc::new(inner),
-        })
+    pub fn new(properties: Option<&spa::utils::dict::DictRef>) -> Result<Self, Error> {
+        // This is a potential "entry point" to the library, so we need to ensure it is initialized.
+        crate::init();
+
+        unsafe {
+            let props = properties
+                .map_or(ptr::null(), |props| props.as_raw())
+                .cast_mut();
+            let l = pw_sys::pw_loop_new(props);
+            let ptr = ptr::NonNull::new(l).ok_or(Error::CreationFailed)?;
+            Ok(Self::from_raw(ptr))
+        }
+    }
+
+    /// Create a new loop from a raw [`pw_loop`](`pw_sys::pw_loop`), taking ownership of it.
+    ///
+    /// # Safety
+    /// The provided pointer must point to a valid, well aligned [`pw_loop`](`pw_sys::pw_loop`).
+    ///
+    /// The raw loop should not be manually destroyed or moved, as the new [`Loop`] takes ownership of it.
+    pub unsafe fn from_raw(ptr: NonNull<pw_sys::pw_loop>) -> Self {
+        Self {
+            inner: Rc::new(LoopInner::from_raw(ptr)),
+        }
     }
 
-    // TODO: fn with_props
     pub fn downgrade(&self) -> WeakLoop {
         let weak = Rc::downgrade(&self.inner);
         WeakLoop { weak }
     }
 }
 
-impl std::convert::AsRef<LoopRef> for Loop {
-    fn as_ref(&self) -> &LoopRef {
-        self.deref()
+// Safety: The inner pw_loop is guaranteed to remain valid while any clone of the `Loop` is held,
+//         because we use an internal Rc to keep it alive.
+unsafe impl IsLoopRc for Loop {}
+
+impl std::ops::Deref for Loop {
+    type Target = LoopRef;
+
+    fn deref(&self) -> &Self::Target {
+        let loop_ = self.inner.ptr.as_ptr();
+        unsafe { &*(loop_.cast::<LoopRef>()) }
     }
 }
 
-impl std::ops::Deref for Loop {
-    type Target = LoopInner;
-
-    fn deref(&self) -> &Self::Target {
-        &self.inner
+impl std::convert::AsRef<LoopRef> for Loop {
+    fn as_ref(&self) -> &LoopRef {
+        self.deref()
     }
 }
 
@@ -414,45 +448,13 @@ impl WeakLoop {
 }
 
 #[derive(Debug)]
-pub struct LoopInner {
+struct LoopInner {
     ptr: ptr::NonNull<pw_sys::pw_loop>,
 }
 
 impl LoopInner {
-    /// Create a new loop from a raw [`pw_loop`](`pw_sys::pw_loop`), taking ownership of it.
-    ///
-    /// # Safety
-    /// The provided pointer must point to a valid, well aligned [`pw_loop`](`pw_sys::pw_loop`).
-    ///
-    /// The raw loop should not be manually destroyed or moved, as the new [`LoopInner`] takes ownership of it.
     pub unsafe fn from_raw(ptr: NonNull<pw_sys::pw_loop>) -> Self {
-        LoopInner { ptr }
-    }
-
-    /// Create a new [`LoopInner`].
-    pub fn new() -> Result<Self, Error> {
-        // This is a potential "entry point" to the library, so we need to ensure it is initialized.
-        crate::init();
-
-        unsafe {
-            let l = pw_sys::pw_loop_new(std::ptr::null());
-            let ptr = ptr::NonNull::new(l).ok_or(Error::CreationFailed)?;
-            Ok(Self::from_raw(ptr))
-        }
-    }
-}
-
-impl std::convert::AsRef<LoopRef> for LoopInner {
-    fn as_ref(&self) -> &LoopRef {
-        self.deref()
-    }
-}
-
-impl std::ops::Deref for LoopInner {
-    type Target = LoopRef;
-
-    fn deref(&self) -> &Self::Target {
-        unsafe { &*(self.ptr.as_ptr() as *mut LoopRef) }
+        Self { ptr }
     }
 }
 
