@@ -3,6 +3,7 @@
 
 use bitflags::bitflags;
 use libc::c_void;
+use std::ops::Deref;
 use std::pin::Pin;
 use std::{ffi::CStr, ptr};
 use std::{fmt, mem};
@@ -11,7 +12,7 @@ use crate::{
     proxy::{Listener, Proxy, ProxyT},
     types::ObjectType,
 };
-use spa::{dict::ForeignDict, pod::Pod, spa_interface_call_method};
+use spa::{pod::Pod, spa_interface_call_method};
 
 #[derive(Debug)]
 pub struct Node {
@@ -111,7 +112,7 @@ impl ProxyT for Node {
 #[derive(Default)]
 struct ListenerLocalCallbacks {
     #[allow(clippy::type_complexity)]
-    info: Option<Box<dyn Fn(&NodeInfo)>>,
+    info: Option<Box<dyn Fn(&NodeInfoRef)>>,
     #[allow(clippy::type_complexity)]
     param: Option<Box<dyn Fn(i32, spa::param::ParamType, u32, u32, Option<&Pod>)>>,
 }
@@ -121,54 +122,48 @@ pub struct NodeListenerLocalBuilder<'a> {
     cbs: ListenerLocalCallbacks,
 }
 
-pub struct NodeInfo {
-    ptr: ptr::NonNull<pw_sys::pw_node_info>,
-    props: Option<ForeignDict>,
-}
+#[repr(transparent)]
+pub struct NodeInfoRef(pw_sys::pw_node_info);
 
-impl NodeInfo {
-    /// Create a `NodeInfo` from a raw `pw_sys::pw_node_info`.
-    ///
-    /// # Safety
-    /// `ptr` must point to a valid, well aligned `pw_sys::pw_node_info`.
-    fn from_raw(ptr: ptr::NonNull<pw_sys::pw_node_info>) -> Self {
-        let props_ptr = unsafe { ptr.as_ref().props };
-        let props = ptr::NonNull::new(props_ptr).map(|ptr| unsafe { ForeignDict::from_ptr(ptr) });
+impl NodeInfoRef {
+    pub fn as_raw(&self) -> &pw_sys::pw_node_info {
+        &self.0
+    }
 
-        Self { ptr, props }
+    pub fn as_raw_ptr(&self) -> *mut pw_sys::pw_node_info {
+        std::ptr::addr_of!(self.0).cast_mut()
     }
 
     pub fn id(&self) -> u32 {
-        unsafe { self.ptr.as_ref().id }
+        self.0.id
     }
 
     pub fn max_input_ports(&self) -> u32 {
-        unsafe { self.ptr.as_ref().max_input_ports }
+        self.0.max_input_ports
     }
 
     pub fn max_output_ports(&self) -> u32 {
-        unsafe { self.ptr.as_ref().max_output_ports }
+        self.0.max_output_ports
     }
 
     pub fn change_mask(&self) -> NodeChangeMask {
-        let mask = unsafe { self.ptr.as_ref().change_mask };
-        NodeChangeMask::from_bits_retain(mask)
+        NodeChangeMask::from_bits_retain(self.0.change_mask)
     }
 
     pub fn n_input_ports(&self) -> u32 {
-        unsafe { self.ptr.as_ref().n_input_ports }
+        self.0.n_input_ports
     }
 
     pub fn n_output_ports(&self) -> u32 {
-        unsafe { self.ptr.as_ref().n_output_ports }
+        self.0.n_output_ports
     }
 
     pub fn state(&self) -> NodeState {
-        let state = unsafe { self.ptr.as_ref().state };
-        match state {
+        let raw_state = self.0.state;
+        match raw_state {
             pw_sys::pw_node_state_PW_NODE_STATE_ERROR => {
                 let error = unsafe {
-                    let error = self.ptr.as_ref().error;
+                    let error = self.0.error;
                     CStr::from_ptr(error).to_str().unwrap()
                 };
                 NodeState::Error(error)
@@ -177,31 +172,92 @@ impl NodeInfo {
             pw_sys::pw_node_state_PW_NODE_STATE_SUSPENDED => NodeState::Suspended,
             pw_sys::pw_node_state_PW_NODE_STATE_IDLE => NodeState::Idle,
             pw_sys::pw_node_state_PW_NODE_STATE_RUNNING => NodeState::Running,
-            _ => panic!("Invalid node state: {}", state),
+            _ => panic!("Invalid node state: {}", raw_state),
         }
     }
 
-    pub fn props(&self) -> Option<&ForeignDict> {
-        self.props.as_ref()
+    pub fn props(&self) -> Option<&spa::utils::dict::DictRef> {
+        let props_ptr: *mut spa::utils::dict::DictRef = self.0.props.cast();
+        ptr::NonNull::new(props_ptr).map(|ptr| unsafe { ptr.as_ref() })
     }
 
     /// Get the param infos for the node.
     pub fn params(&self) -> &[spa::param::ParamInfo] {
         unsafe {
-            let params = self.ptr.as_ref().params;
+            let params_ptr = self.0.params;
 
-            if params.is_null() {
+            if params_ptr.is_null() {
                 &[]
             } else {
                 std::slice::from_raw_parts(
-                    params as *const _,
-                    self.ptr.as_ref().n_params.try_into().unwrap(),
+                    params_ptr as *const _,
+                    self.0.n_params.try_into().unwrap(),
                 )
             }
         }
     }
 }
 
+impl fmt::Debug for NodeInfoRef {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+        f.debug_struct("NodeInfoRef")
+            .field("id", &self.id())
+            .field("max-input-ports", &self.max_input_ports())
+            .field("max-output-ports", &self.max_output_ports())
+            .field("change-mask", &self.change_mask())
+            .field("n-input-ports", &self.n_input_ports())
+            .field("n-output-ports", &self.n_output_ports())
+            .field("state", &self.state())
+            .field("props", &self.props())
+            .field("params", &self.params())
+            .finish()
+    }
+}
+
+pub struct NodeInfo {
+    ptr: ptr::NonNull<pw_sys::pw_node_info>,
+}
+
+impl NodeInfo {
+    pub fn new(ptr: ptr::NonNull<pw_sys::pw_node_info>) -> Self {
+        Self { ptr }
+    }
+
+    /// Create a `NodeInfo` from a raw `pw_sys::pw_node_info`.
+    ///
+    /// # Safety
+    /// `ptr` must point to a valid, well aligned `pw_sys::pw_node_info`.
+    pub fn from_raw(raw: *mut pw_sys::pw_node_info) -> Self {
+        Self {
+            ptr: ptr::NonNull::new(raw).expect("Provided pointer is null"),
+        }
+    }
+
+    pub fn into_raw(self) -> *mut pw_sys::pw_node_info {
+        std::mem::ManuallyDrop::new(self).ptr.as_ptr()
+    }
+}
+
+impl Drop for NodeInfo {
+    fn drop(&mut self) {
+        unsafe { pw_sys::pw_node_info_free(self.ptr.as_ptr()) }
+    }
+}
+
+impl std::ops::Deref for NodeInfo {
+    type Target = NodeInfoRef;
+
+    fn deref(&self) -> &Self::Target {
+        unsafe { self.ptr.cast::<NodeInfoRef>().as_ref() }
+    }
+}
+
+impl AsRef<NodeInfoRef> for NodeInfo {
+    fn as_ref(&self) -> &NodeInfoRef {
+        self.deref()
+    }
+}
+
 bitflags! {
     #[derive(Debug, PartialEq, Eq, Clone, Copy)]
     pub struct NodeChangeMask: u64 {
@@ -251,7 +307,7 @@ impl Listener for NodeListener {}
 
 impl Drop for NodeListener {
     fn drop(&mut self) {
-        spa::hook::remove(*self.listener);
+        spa::utils::hook::remove(*self.listener);
     }
 }
 
@@ -259,7 +315,7 @@ impl<'a> NodeListenerLocalBuilder<'a> {
     #[must_use]
     pub fn info<F>(mut self, info: F) -> Self
     where
-        F: Fn(&NodeInfo) + 'static,
+        F: Fn(&NodeInfoRef) + 'static,
     {
         self.cbs.info = Some(Box::new(info));
         self
@@ -281,9 +337,9 @@ impl<'a> NodeListenerLocalBuilder<'a> {
             info: *const pw_sys::pw_node_info,
         ) {
             let callbacks = (data as *mut ListenerLocalCallbacks).as_ref().unwrap();
-            let info = ptr::NonNull::new(info as *mut _).expect("info is NULL");
-            let info = NodeInfo::from_raw(info);
-            callbacks.info.as_ref().unwrap()(&info);
+            let info = ptr::NonNull::new(info as *mut pw_sys::pw_node_info).expect("info is NULL");
+            let info = info.cast::<NodeInfoRef>().as_ref();
+            callbacks.info.as_ref().unwrap()(info);
         }
 
         unsafe extern "C" fn node_events_param(
