@@ -1,21 +1,22 @@
 // Copyright The pipewire-rs Contributors.
 // SPDX-License-Identifier: MIT
 
-use bitflags::bitflags;
 use libc::{c_char, c_void};
-use std::mem;
-use std::pin::Pin;
+
 use std::{
     ffi::{CStr, CString},
+    mem,
+    pin::Pin,
     ptr,
 };
 
 use crate::{
+    permissions::PermissionFlags,
+    properties::Properties,
     proxy::{Proxy, ProxyT},
     types::ObjectType,
-    Error, Properties,
+    Error,
 };
-use spa::{dict::ForeignDict, prelude::*};
 
 #[derive(Debug)]
 pub struct Registry {
@@ -40,7 +41,10 @@ impl Registry {
         }
     }
 
-    pub fn bind<T: ProxyT, D: ReadableDict>(&self, object: &GlobalObject<D>) -> Result<T, Error> {
+    pub fn bind<T: ProxyT, P: AsRef<spa::utils::dict::DictRef>>(
+        &self,
+        object: &GlobalObject<P>,
+    ) -> Result<T, Error> {
         let proxy = unsafe {
             let type_ = CString::new(object.type_.to_str()).unwrap();
             let version = object.type_.client_version();
@@ -64,7 +68,7 @@ impl Registry {
     }
 
     /// Attempt to destroy the global object with the specified id on the remote.
-    pub fn destroy_global(&self, global_id: u32) -> spa::SpaResult {
+    pub fn destroy_global(&self, global_id: u32) -> spa::utils::result::SpaResult {
         let result = unsafe {
             spa::spa_interface_call_method!(
                 self.as_ptr(),
@@ -74,7 +78,7 @@ impl Registry {
             )
         };
 
-        spa::SpaResult::from_c(result)
+        spa::utils::result::SpaResult::from_c(result)
     }
 }
 
@@ -86,11 +90,13 @@ impl Drop for Registry {
     }
 }
 
+type GlobalCallback = dyn Fn(&GlobalObject<&spa::utils::dict::DictRef>);
+type GlobalRemoveCallback = dyn Fn(u32);
+
 #[derive(Default)]
 struct ListenerLocalCallbacks {
-    #[allow(clippy::type_complexity)]
-    global: Option<Box<dyn Fn(&GlobalObject<ForeignDict>)>>,
-    global_remove: Option<Box<dyn Fn(u32)>>,
+    global: Option<Box<GlobalCallback>>,
+    global_remove: Option<Box<GlobalRemoveCallback>>,
 }
 
 pub struct ListenerLocalBuilder<'a> {
@@ -109,7 +115,7 @@ pub struct Listener {
 
 impl Drop for Listener {
     fn drop(&mut self) {
-        spa::hook::remove(*self.listener);
+        spa::utils::hook::remove(*self.listener);
     }
 }
 
@@ -117,7 +123,7 @@ impl<'a> ListenerLocalBuilder<'a> {
     #[must_use]
     pub fn global<F>(mut self, global: F) -> Self
     where
-        F: Fn(&GlobalObject<ForeignDict>) + 'static,
+        F: Fn(&GlobalObject<&spa::utils::dict::DictRef>) + 'static,
     {
         self.cbs.global = Some(Box::new(global));
         self
@@ -193,29 +199,17 @@ impl<'a> ListenerLocalBuilder<'a> {
     }
 }
 
-bitflags! {
-    #[derive(Debug, PartialEq, Eq, Clone, Copy)]
-    pub struct Permission: u32 {
-        const R = pw_sys::PW_PERM_R;
-        const W = pw_sys::PW_PERM_W;
-        const X = pw_sys::PW_PERM_X;
-        const M = pw_sys::PW_PERM_M;
-        #[cfg(feature = "v0_3_77")]
-        const L = pw_sys::PW_PERM_L;
-    }
-}
-
 #[derive(Debug)]
-pub struct GlobalObject<D: ReadableDict> {
+pub struct GlobalObject<P: AsRef<spa::utils::dict::DictRef>> {
     pub id: u32,
-    pub permissions: Permission,
+    pub permissions: PermissionFlags,
     pub type_: ObjectType,
     pub version: u32,
-    pub props: Option<D>,
+    pub props: Option<P>,
 }
 
-impl GlobalObject<ForeignDict> {
-    fn new(
+impl GlobalObject<&spa::utils::dict::DictRef> {
+    unsafe fn new(
         id: u32,
         permissions: u32,
         type_: &str,
@@ -223,9 +217,9 @@ impl GlobalObject<ForeignDict> {
         props: *const spa_sys::spa_dict,
     ) -> Self {
         let type_ = ObjectType::from_str(type_);
-        let permissions = Permission::from_bits_retain(permissions);
-        let props = props as *mut _;
-        let props = ptr::NonNull::new(props).map(|ptr| unsafe { ForeignDict::from_ptr(ptr) });
+        let permissions = PermissionFlags::from_bits_retain(permissions);
+        let props = ptr::NonNull::new(props.cast_mut())
+            .map(|ptr| ptr.cast::<spa::utils::dict::DictRef>().as_ref());
 
         Self {
             id,
@@ -237,7 +231,7 @@ impl GlobalObject<ForeignDict> {
     }
 }
 
-impl<D: ReadableDict> GlobalObject<D> {
+impl<P: AsRef<spa::utils::dict::DictRef>> GlobalObject<P> {
     pub fn to_owned(&self) -> GlobalObject<Properties> {
         GlobalObject {
             id: self.id,
@@ -247,7 +241,7 @@ impl<D: ReadableDict> GlobalObject<D> {
             props: self
                 .props
                 .as_ref()
-                .map(|props| Properties::from_dict(props)),
+                .map(|props| Properties::from_dict(props.as_ref())),
         }
     }
 }
