@@ -4,18 +4,22 @@
 //! Pipewire Stream
 
 use crate::buffer::Buffer;
-use crate::{error::Error, Core, MainLoop, Properties, PropertiesRef};
+use crate::{
+    core::Core,
+    error::Error,
+    properties::{Properties, PropertiesRef},
+};
 use bitflags::bitflags;
-use spa::result::SpaResult;
-use std::fmt::Debug;
+use spa::utils::result::SpaResult;
 use std::{
     ffi::{self, CStr, CString},
+    fmt::Debug,
     mem, os,
     pin::Pin,
     ptr,
 };
 
-#[derive(Debug)]
+#[derive(Debug, PartialEq)]
 pub enum StreamState {
     Error(String),
     Unconnected,
@@ -46,111 +50,103 @@ impl StreamState {
 
 /// A wrapper around the pipewire stream interface. Streams are a higher
 /// level abstraction around nodes in the graph. A stream can be used to send or
-/// receive frames of audio of video data by connecting it to another node.
+/// receive frames of audio or video data by connecting it to another node.
 /// `D` is the user data, to allow passing extra context to the callbacks.
-pub struct Stream<D> {
+pub struct Stream {
     ptr: ptr::NonNull<pw_sys::pw_stream>,
     // objects that need to stay alive while the Stream is
-    _alive: KeepAlive<D>,
+    _core: Core,
 }
 
-enum KeepAlive<D> {
-    // Stream created with Stream::new()
-    Normal {
-        _core: Core,
-    },
-    // Stream created with Stream::simple()
-    Simple {
-        _events: Pin<Box<pw_sys::pw_stream_events>>,
-        _data: Box<ListenerLocalCallbacks<D>>,
-    },
-    // Temporary stream for callbacks
-    Temp,
-}
-
-impl<D> Stream<D> {
+impl Stream {
     /// Create a [`Stream`]
     ///
     /// Initialises a new stream with the given `name` and `properties`.
     pub fn new(core: &Core, name: &str, properties: Properties) -> Result<Self, Error> {
         let name = CString::new(name).expect("Invalid byte in stream name");
-        let stream =
-            unsafe { pw_sys::pw_stream_new(core.as_ptr(), name.as_ptr(), properties.into_raw()) };
+        let stream = unsafe {
+            pw_sys::pw_stream_new(core.as_raw_ptr(), name.as_ptr(), properties.into_raw())
+        };
         let stream = ptr::NonNull::new(stream).ok_or(Error::CreationFailed)?;
 
         Ok(Stream {
             ptr: stream,
-            _alive: KeepAlive::Normal {
-                _core: core.clone(),
-            },
+            _core: core.clone(),
         })
     }
 
-    /// Create a [`Stream`] with custom user data, and connect its event.
-    ///
-    /// Create a stream directly from a [`MainLoop`]. This avoids having to create
-    /// a [`crate::Context`] and [`Core`] yourself in cases that don't require anything
-    /// special.
-    ///
-    /// # Panics
-    /// Will panic if `name` contains a 0 byte.
-    ///
-    /// # Example
-    /// ```no_run
-    /// use pipewire::prelude::*;
-    /// use pipewire::properties;
-    ///
-    /// let mainloop = pipewire::MainLoop::new()?;
-    ///
-    /// let mut stream = pipewire::stream::Stream::<i32>::with_user_data(
-    ///     &mainloop,
-    ///     "video-test",
-    ///     properties! {
-    ///         *pipewire::keys::MEDIA_TYPE => "Video",
-    ///         *pipewire::keys::MEDIA_CATEGORY => "Capture",
-    ///         *pipewire::keys::MEDIA_ROLE => "Camera",
-    ///     },
-    ///     42,
-    /// )
-    /// .state_changed(|old, new| {
-    ///     println!("State changed: {:?} -> {:?}", old, new);
-    /// })
-    /// .process(|_stream, _user_data| {
-    ///     println!("On frame");
-    /// })
-    /// .create()?;
-    /// # Ok::<(), pipewire::Error>(())
-    /// ```
-    pub fn with_user_data<'a>(
-        main_loop: &'a MainLoop,
-        name: &str,
-        properties: Properties,
-        user_data: D,
-    ) -> SimpleLocalBuilder<'a, D> {
-        let name = CString::new(name).expect("Invalid byte in stream name");
+    pub fn into_raw(self) -> *mut pw_sys::pw_stream {
+        let mut this = std::mem::ManuallyDrop::new(self);
 
-        SimpleLocalBuilder::<D> {
-            main_loop,
-            name,
-            properties,
-            callbacks: ListenerLocalCallbacks::with_user_data(user_data),
+        // FIXME: self needs to be wrapped in ManuallyDrop so the raw stream
+        //        isn't destroyed. However, the core should still be dropped.
+        //        Is there a cleaner and safer way to drop the core than like this?
+        unsafe {
+            ptr::drop_in_place(ptr::addr_of_mut!(this._core));
         }
+
+        this.ptr.as_ptr()
+    }
+}
+
+impl std::ops::Deref for Stream {
+    type Target = StreamRef;
+
+    fn deref(&self) -> &Self::Target {
+        unsafe { self.ptr.cast().as_ref() }
+    }
+}
+
+impl std::fmt::Debug for Stream {
+    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
+        f.debug_struct("Stream")
+            .field("name", &self.name())
+            .field("state", &self.state())
+            .field("node-id", &self.node_id())
+            .field("properties", &self.properties())
+            .finish()
+    }
+}
+
+impl std::ops::Drop for Stream {
+    fn drop(&mut self) {
+        unsafe { pw_sys::pw_stream_destroy(self.as_raw_ptr()) }
+    }
+}
+
+#[repr(transparent)]
+pub struct StreamRef(pw_sys::pw_stream);
+
+impl StreamRef {
+    pub fn as_raw(&self) -> &pw_sys::pw_stream {
+        &self.0
+    }
+
+    pub fn as_raw_ptr(&self) -> *mut pw_sys::pw_stream {
+        ptr::addr_of!(self.0).cast_mut()
     }
 
     /// Add a local listener builder
     #[must_use = "Fluent builder API"]
-    pub fn add_local_listener_with_user_data(
-        &mut self,
+    pub fn add_local_listener_with_user_data<D>(
+        &self,
         user_data: D,
     ) -> ListenerLocalBuilder<'_, D> {
         let mut callbacks = ListenerLocalCallbacks::with_user_data(user_data);
-        callbacks.stream = Some(self.ptr);
+        callbacks.stream =
+            Some(ptr::NonNull::new(self.as_raw_ptr()).expect("Pointer should be nonnull"));
         ListenerLocalBuilder {
             stream: self,
             callbacks,
         }
     }
 
+    /// Add a local listener builder. User data is initialized with its default value
+    #[must_use = "Fluent builder API"]
+    pub fn add_local_listener<D: Default>(&self) -> ListenerLocalBuilder<'_, D> {
+        self.add_local_listener_with_user_data(Default::default())
+    }
+
     /// Connect the stream
     ///
     /// Tries to connect to the node `id` in the given `direction`. If no node
@@ -158,18 +154,20 @@ impl<D> Stream<D> {
     // FIXME: high-level API for params
     pub fn connect(
         &self,
-        direction: spa::Direction,
+        direction: spa::utils::Direction,
         id: Option<u32>,
         flags: StreamFlags,
-        params: &mut [*const spa_sys::spa_pod],
+        params: &mut [&spa::pod::Pod],
     ) -> Result<(), Error> {
         let r = unsafe {
             pw_sys::pw_stream_connect(
-                self.as_ptr(),
+                self.as_raw_ptr(),
                 direction.as_raw(),
                 id.unwrap_or(crate::constants::ID_ANY),
                 flags.bits(),
-                params.as_mut_ptr(),
+                // We cast from *mut [&spa::pod::Pod] to *mut [*const spa_sys::spa_pod] here,
+                // which is valid because spa::pod::Pod is a transparent wrapper around spa_sys::spa_pod
+                params.as_mut_ptr().cast(),
                 params.len() as u32,
             )
         };
@@ -183,9 +181,13 @@ impl<D> Stream<D> {
     /// Call from the `param_changed` callback to negotiate a new set of
     /// parameters for the stream.
     // FIXME: high-level API for params
-    pub fn update_params(&self, params: &mut [*const spa_sys::spa_pod]) -> Result<(), Error> {
+    pub fn update_params(&self, params: &mut [&spa::pod::Pod]) -> Result<(), Error> {
         let r = unsafe {
-            pw_sys::pw_stream_update_params(self.as_ptr(), params.as_mut_ptr(), params.len() as u32)
+            pw_sys::pw_stream_update_params(
+                self.as_raw_ptr(),
+                params.as_mut_ptr().cast(),
+                params.len() as u32,
+            )
         };
 
         SpaResult::from_c(r).into_sync_result()?;
@@ -194,7 +196,7 @@ impl<D> Stream<D> {
 
     /// Activate or deactivate the stream
     pub fn set_active(&self, active: bool) -> Result<(), Error> {
-        let r = unsafe { pw_sys::pw_stream_set_active(self.as_ptr(), active) };
+        let r = unsafe { pw_sys::pw_stream_set_active(self.as_raw_ptr(), active) };
 
         SpaResult::from_c(r).into_sync_result()?;
         Ok(())
@@ -211,10 +213,10 @@ impl<D> Stream<D> {
     /// The pointer returned could be NULL if no buffer is available. The buffer
     /// should be returned to the stream once processing is complete.
     pub unsafe fn dequeue_raw_buffer(&self) -> *mut pw_sys::pw_buffer {
-        pw_sys::pw_stream_dequeue_buffer(self.as_ptr())
+        pw_sys::pw_stream_dequeue_buffer(self.as_raw_ptr())
     }
 
-    pub fn dequeue_buffer(&self) -> Option<Buffer<D>> {
+    pub fn dequeue_buffer(&self) -> Option<Buffer> {
         unsafe { Buffer::from_raw(self.dequeue_raw_buffer(), self) }
     }
 
@@ -227,18 +229,14 @@ impl<D> Stream<D> {
     /// # Safety
     ///
     /// The buffer pointer should be one obtained from this stream instance by
-    /// a call to [Stream::dequeue_raw_buffer()].
+    /// a call to [StreamRef::dequeue_raw_buffer()].
     pub unsafe fn queue_raw_buffer(&self, buffer: *mut pw_sys::pw_buffer) {
-        pw_sys::pw_stream_queue_buffer(self.as_ptr(), buffer);
-    }
-
-    fn as_ptr(&self) -> *mut pw_sys::pw_stream {
-        self.ptr.as_ptr()
+        pw_sys::pw_stream_queue_buffer(self.as_raw_ptr(), buffer);
     }
 
     /// Disconnect the stream
     pub fn disconnect(&self) -> Result<(), Error> {
-        let r = unsafe { pw_sys::pw_stream_disconnect(self.as_ptr()) };
+        let r = unsafe { pw_sys::pw_stream_disconnect(self.as_raw_ptr()) };
 
         SpaResult::from_c(r).into_sync_result()?;
         Ok(())
@@ -252,27 +250,38 @@ impl<D> Stream<D> {
     pub fn set_error(&mut self, res: i32, error: &str) {
         let error = CString::new(error).expect("failed to convert error to CString");
         unsafe {
-            pw_sys::pw_stream_set_error(self.as_ptr(), res, error.as_c_str().as_ptr());
+            pw_sys::pw_stream_set_error(self.as_raw_ptr(), res, error.as_c_str().as_ptr());
         }
     }
 
     /// Flush the stream. When  `drain` is `true`, the `drained` callback will
     /// be called when all data is played or recorded.
     pub fn flush(&self, drain: bool) -> Result<(), Error> {
-        let r = unsafe { pw_sys::pw_stream_flush(self.as_ptr(), drain) };
+        let r = unsafe { pw_sys::pw_stream_flush(self.as_raw_ptr(), drain) };
 
         SpaResult::from_c(r).into_sync_result()?;
         Ok(())
     }
 
-    // TODO: pw_stream_set_control()
+    pub fn set_control(&self, id: u32, values: &[f32]) -> Result<(), Error> {
+        let r = unsafe {
+            pw_sys::pw_stream_set_control(
+                self.as_raw_ptr(),
+                id,
+                values.len() as u32,
+                values.as_ptr() as *mut f32,
+            )
+        };
+        SpaResult::from_c(r).into_sync_result()?;
+        Ok(())
+    }
 
     // getters
 
     /// Get the name of the stream.
     pub fn name(&self) -> String {
         let name = unsafe {
-            let name = pw_sys::pw_stream_get_name(self.as_ptr());
+            let name = pw_sys::pw_stream_get_name(self.as_raw_ptr());
             CStr::from_ptr(name)
         };
 
@@ -282,108 +291,71 @@ impl<D> Stream<D> {
     /// Get the current state of the stream.
     pub fn state(&self) -> StreamState {
         let mut error: *const std::os::raw::c_char = ptr::null();
-        let state =
-            unsafe { pw_sys::pw_stream_get_state(self.as_ptr(), (&mut error) as *mut *const _) };
+        let state = unsafe {
+            pw_sys::pw_stream_get_state(self.as_raw_ptr(), (&mut error) as *mut *const _)
+        };
         StreamState::from_raw(state, error)
     }
 
     /// Get the properties of the stream.
-    pub fn properties(&self) -> PropertiesRef<'_> {
+    pub fn properties(&self) -> &PropertiesRef {
         unsafe {
-            let props = pw_sys::pw_stream_get_properties(self.as_ptr());
-            let props = ptr::NonNull::new(props as *mut _).expect("stream properties is NULL");
-            PropertiesRef::from_ptr(props)
+            let props = pw_sys::pw_stream_get_properties(self.as_raw_ptr());
+            let props = ptr::NonNull::new(props.cast_mut()).expect("stream properties is NULL");
+            props.cast().as_ref()
         }
     }
 
     /// Get the node ID of the stream.
     pub fn node_id(&self) -> u32 {
-        unsafe { pw_sys::pw_stream_get_node_id(self.as_ptr()) }
+        unsafe { pw_sys::pw_stream_get_node_id(self.as_raw_ptr()) }
+    }
+
+    #[cfg(feature = "v0_3_34")]
+    pub fn is_driving(&self) -> bool {
+        unsafe { pw_sys::pw_stream_is_driving(self.as_raw_ptr()) }
+    }
+
+    #[cfg(feature = "v0_3_34")]
+    pub fn trigger_process(&self) -> Result<(), Error> {
+        let r = unsafe { pw_sys::pw_stream_trigger_process(self.as_raw_ptr()) };
+
+        SpaResult::from_c(r).into_result()?;
+        Ok(())
     }
 
     // TODO: pw_stream_get_core()
     // TODO: pw_stream_get_time()
 }
 
-impl<D: Default> Stream<D> {
-    /// Create a [`Stream`] and connect its event.
-    ///
-    /// Create a stream directly from a [`MainLoop`]. This avoids having to create
-    /// a [`crate::Context`] and [`Core`] yourself in cases that don't require anything
-    /// special.
-    ///
-    /// # Panics
-    /// Will panic if `name` contains a 0 byte.
-    ///
-    /// # Example
-    /// ```no_run
-    /// use pipewire::prelude::*;
-    /// use pipewire::properties;
-    ///
-    /// let mainloop = pipewire::MainLoop::new()?;
-    ///
-    /// let mut stream = pipewire::stream::Stream::<()>::simple(
-    ///     &mainloop,
-    ///     "video-test",
-    ///     properties! {
-    ///         *pipewire::keys::MEDIA_TYPE => "Video",
-    ///         *pipewire::keys::MEDIA_CATEGORY => "Capture",
-    ///         *pipewire::keys::MEDIA_ROLE => "Camera",
-    ///     },
-    /// )
-    /// .state_changed(|old, new| {
-    ///     println!("State changed: {:?} -> {:?}", old, new);
-    /// })
-    /// .process(|_stream, _user_data| {
-    ///     println!("On frame");
-    /// })
-    /// .create()?;
-    /// # Ok::<(), pipewire::Error>(())
-    /// ```
-    #[must_use]
-    pub fn simple<'a>(
-        main_loop: &'a MainLoop,
-        name: &str,
-        properties: Properties,
-    ) -> SimpleLocalBuilder<'a, D> {
-        Self::with_user_data(main_loop, name, properties, Default::default())
-    }
-
-    /// Add a local listener builder
-    #[must_use = "Fluent builder API"]
-    pub fn add_local_listener(&mut self) -> ListenerLocalBuilder<'_, D> {
-        self.add_local_listener_with_user_data(Default::default())
-    }
-}
-
-impl<D> std::fmt::Debug for Stream<D> {
-    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
-        f.debug_struct("Stream")
-            .field("name", &self.name())
-            .field("state", &self.state())
-            .field("node-id", &self.node_id())
-            .field("properties", &self.properties())
-            .finish()
-    }
-}
-
-type ParamChangedCB<D> = dyn Fn(u32, &mut D, *const spa_sys::spa_pod);
-type ProcessCB<D> = dyn Fn(&Stream<D>, &mut D);
+type ParamChangedCB<D> = dyn FnMut(&StreamRef, &mut D, u32, Option<&spa::pod::Pod>);
+type ProcessCB<D> = dyn FnMut(&StreamRef, &mut D);
 
+#[allow(clippy::type_complexity)]
 pub struct ListenerLocalCallbacks<D> {
-    pub state_changed: Option<Box<dyn Fn(StreamState, StreamState)>>,
-    pub control_info: Option<Box<dyn Fn(u32, *const pw_sys::pw_stream_control)>>,
-    #[allow(clippy::type_complexity)]
-    pub io_changed: Option<Box<dyn Fn(u32, *mut os::raw::c_void, u32)>>,
+    pub state_changed: Option<Box<dyn FnMut(&StreamRef, &mut D, StreamState, StreamState)>>,
+    pub control_info:
+        Option<Box<dyn FnMut(&StreamRef, &mut D, u32, *const pw_sys::pw_stream_control)>>,
+    pub io_changed: Option<Box<dyn FnMut(&StreamRef, &mut D, u32, *mut os::raw::c_void, u32)>>,
     pub param_changed: Option<Box<ParamChangedCB<D>>>,
-    pub add_buffer: Option<Box<dyn Fn(*mut pw_sys::pw_buffer)>>,
-    pub remove_buffer: Option<Box<dyn Fn(*mut pw_sys::pw_buffer)>>,
+    pub add_buffer: Option<Box<dyn FnMut(&StreamRef, &mut D, *mut pw_sys::pw_buffer)>>,
+    pub remove_buffer: Option<Box<dyn FnMut(&StreamRef, &mut D, *mut pw_sys::pw_buffer)>>,
     pub process: Option<Box<ProcessCB<D>>>,
-    pub drained: Option<Box<dyn Fn()>>,
+    pub drained: Option<Box<dyn FnMut(&StreamRef, &mut D)>>,
+    #[cfg(feature = "v0_3_39")]
+    pub command: Option<Box<dyn FnMut(&StreamRef, &mut D, *const spa_sys::spa_command)>>,
+    #[cfg(feature = "v0_3_40")]
+    pub trigger_done: Option<Box<dyn FnMut(&StreamRef, &mut D)>>,
     pub user_data: D,
     stream: Option<ptr::NonNull<pw_sys::pw_stream>>,
 }
 
+unsafe fn unwrap_stream_ptr<'a>(stream: Option<ptr::NonNull<pw_sys::pw_stream>>) -> &'a StreamRef {
+    stream
+        .map(|ptr| ptr.cast::<StreamRef>().as_ref())
+        .expect("stream cannot be null")
+}
+
 impl<D> ListenerLocalCallbacks<D> {
     fn with_user_data(user_data: D) -> Self {
         ListenerLocalCallbacks {
@@ -396,6 +368,10 @@ impl<D> ListenerLocalCallbacks<D> {
             param_changed: Default::default(),
             remove_buffer: Default::default(),
             state_changed: Default::default(),
+            #[cfg(feature = "v0_3_39")]
+            command: Default::default(),
+            #[cfg(feature = "v0_3_40")]
+            trigger_done: Default::default(),
             user_data,
         }
     }
@@ -414,11 +390,12 @@ impl<D> ListenerLocalCallbacks<D> {
             new: pw_sys::pw_stream_state,
             error: *const os::raw::c_char,
         ) {
-            if let Some(state) = (data as *mut ListenerLocalCallbacks<D>).as_ref() {
-                if let Some(ref cb) = state.state_changed {
+            if let Some(state) = (data as *mut ListenerLocalCallbacks<D>).as_mut() {
+                if let Some(cb) = &mut state.state_changed {
+                    let stream = unwrap_stream_ptr(state.stream);
                     let old = StreamState::from_raw(old, error);
                     let new = StreamState::from_raw(new, error);
-                    cb(old, new)
+                    cb(stream, &mut state.user_data, old, new)
                 };
             }
         }
@@ -428,9 +405,10 @@ impl<D> ListenerLocalCallbacks<D> {
             id: u32,
             control: *const pw_sys::pw_stream_control,
         ) {
-            if let Some(state) = (data as *mut ListenerLocalCallbacks<D>).as_ref() {
-                if let Some(ref cb) = state.control_info {
-                    cb(id, control);
+            if let Some(state) = (data as *mut ListenerLocalCallbacks<D>).as_mut() {
+                if let Some(cb) = &mut state.control_info {
+                    let stream = unwrap_stream_ptr(state.stream);
+                    cb(stream, &mut state.user_data, id, control);
                 }
             }
         }
@@ -441,9 +419,10 @@ impl<D> ListenerLocalCallbacks<D> {
             area: *mut os::raw::c_void,
             size: u32,
         ) {
-            if let Some(state) = (data as *mut ListenerLocalCallbacks<D>).as_ref() {
-                if let Some(ref cb) = state.io_changed {
-                    cb(id, area, size);
+            if let Some(state) = (data as *mut ListenerLocalCallbacks<D>).as_mut() {
+                if let Some(cb) = &mut state.io_changed {
+                    let stream = unwrap_stream_ptr(state.stream);
+                    cb(stream, &mut state.user_data, id, area, size);
                 }
             }
         }
@@ -454,8 +433,15 @@ impl<D> ListenerLocalCallbacks<D> {
             param: *const spa_sys::spa_pod,
         ) {
             if let Some(state) = (data as *mut ListenerLocalCallbacks<D>).as_mut() {
-                if let Some(ref cb) = state.param_changed {
-                    cb(id, &mut state.user_data, param);
+                if let Some(cb) = &mut state.param_changed {
+                    let stream = unwrap_stream_ptr(state.stream);
+                    let param = if !param.is_null() {
+                        Some(spa::pod::Pod::from_raw(param))
+                    } else {
+                        None
+                    };
+
+                    cb(stream, &mut state.user_data, id, param);
                 }
             }
         }
@@ -464,9 +450,10 @@ impl<D> ListenerLocalCallbacks<D> {
             data: *mut ::std::os::raw::c_void,
             buffer: *mut pw_sys::pw_buffer,
         ) {
-            if let Some(state) = (data as *mut ListenerLocalCallbacks<D>).as_ref() {
-                if let Some(ref cb) = state.add_buffer {
-                    cb(buffer);
+            if let Some(state) = (data as *mut ListenerLocalCallbacks<D>).as_mut() {
+                if let Some(cb) = &mut state.add_buffer {
+                    let stream = unwrap_stream_ptr(state.stream);
+                    cb(stream, &mut state.user_data, buffer);
                 }
             }
         }
@@ -475,32 +462,51 @@ impl<D> ListenerLocalCallbacks<D> {
             data: *mut ::std::os::raw::c_void,
             buffer: *mut pw_sys::pw_buffer,
         ) {
-            if let Some(state) = (data as *mut ListenerLocalCallbacks<D>).as_ref() {
-                if let Some(ref cb) = state.remove_buffer {
-                    cb(buffer);
+            if let Some(state) = (data as *mut ListenerLocalCallbacks<D>).as_mut() {
+                if let Some(cb) = &mut state.remove_buffer {
+                    let stream = unwrap_stream_ptr(state.stream);
+                    cb(stream, &mut state.user_data, buffer);
                 }
             }
         }
 
         unsafe extern "C" fn on_process<D>(data: *mut ::std::os::raw::c_void) {
             if let Some(state) = (data as *mut ListenerLocalCallbacks<D>).as_mut() {
-                if let Some(ref cb) = state.process {
-                    let stream = state
-                        .stream
-                        .map(|ptr| Stream {
-                            ptr,
-                            _alive: KeepAlive::Temp,
-                        })
-                        .expect("stream cannot be null");
-                    cb(&stream, &mut state.user_data);
+                if let Some(cb) = &mut state.process {
+                    let stream = unwrap_stream_ptr(state.stream);
+                    cb(stream, &mut state.user_data);
                 }
             }
         }
 
         unsafe extern "C" fn on_drained<D>(data: *mut ::std::os::raw::c_void) {
-            if let Some(state) = (data as *mut ListenerLocalCallbacks<D>).as_ref() {
-                if let Some(ref cb) = state.drained {
-                    cb();
+            if let Some(state) = (data as *mut ListenerLocalCallbacks<D>).as_mut() {
+                if let Some(cb) = &mut state.drained {
+                    let stream = unwrap_stream_ptr(state.stream);
+                    cb(stream, &mut state.user_data);
+                }
+            }
+        }
+
+        #[cfg(feature = "v0_3_39")]
+        unsafe extern "C" fn on_command<D>(
+            data: *mut ::std::os::raw::c_void,
+            command: *const spa_sys::spa_command,
+        ) {
+            if let Some(state) = (data as *mut ListenerLocalCallbacks<D>).as_mut() {
+                if let Some(cb) = &mut state.command {
+                    let stream = unwrap_stream_ptr(state.stream);
+                    cb(stream, &mut state.user_data, command);
+                }
+            }
+        }
+
+        #[cfg(feature = "v0_3_40")]
+        unsafe extern "C" fn on_trigger_done<D>(data: *mut ::std::os::raw::c_void) {
+            if let Some(state) = (data as *mut ListenerLocalCallbacks<D>).as_mut() {
+                if let Some(cb) = &mut state.trigger_done {
+                    let stream = unwrap_stream_ptr(state.stream);
+                    cb(stream, &mut state.user_data);
                 }
             }
         }
@@ -533,6 +539,14 @@ impl<D> ListenerLocalCallbacks<D> {
             if callbacks.drained.is_some() {
                 events.drained = Some(on_drained::<D>);
             }
+            #[cfg(feature = "v0_3_39")]
+            if callbacks.command.is_some() {
+                events.command = Some(on_command::<D>);
+            }
+            #[cfg(feature = "v0_3_40")]
+            if callbacks.trigger_done.is_some() {
+                events.trigger_done = Some(on_trigger_done::<D>);
+            }
 
             events
         };
@@ -541,94 +555,85 @@ impl<D> ListenerLocalCallbacks<D> {
     }
 }
 
-pub trait ListenerBuilderT<D>: Sized {
-    fn callbacks(&mut self) -> &mut ListenerLocalCallbacks<D>;
+#[must_use]
+pub struct ListenerLocalBuilder<'a, D> {
+    stream: &'a StreamRef,
+    callbacks: ListenerLocalCallbacks<D>,
+}
 
+impl<'a, D> ListenerLocalBuilder<'a, D> {
     /// Set the callback for the `state_changed` event.
-    fn state_changed<F>(mut self, callback: F) -> Self
+    pub fn state_changed<F>(mut self, callback: F) -> Self
     where
-        F: Fn(StreamState, StreamState) + 'static,
+        F: FnMut(&StreamRef, &mut D, StreamState, StreamState) + 'static,
     {
-        self.callbacks().state_changed = Some(Box::new(callback));
+        self.callbacks.state_changed = Some(Box::new(callback));
         self
     }
 
     /// Set the callback for the `control_info` event.
-    fn control_info<F>(mut self, callback: F) -> Self
+    pub fn control_info<F>(mut self, callback: F) -> Self
     where
-        F: Fn(u32, *const pw_sys::pw_stream_control) + 'static,
+        F: FnMut(&StreamRef, &mut D, u32, *const pw_sys::pw_stream_control) + 'static,
     {
-        self.callbacks().control_info = Some(Box::new(callback));
+        self.callbacks.control_info = Some(Box::new(callback));
         self
     }
 
     /// Set the callback for the `io_changed` event.
-    fn io_changed<F>(mut self, callback: F) -> Self
+    pub fn io_changed<F>(mut self, callback: F) -> Self
     where
-        F: Fn(u32, *mut os::raw::c_void, u32) + 'static,
+        F: FnMut(&StreamRef, &mut D, u32, *mut os::raw::c_void, u32) + 'static,
     {
-        self.callbacks().io_changed = Some(Box::new(callback));
+        self.callbacks.io_changed = Some(Box::new(callback));
         self
     }
 
     /// Set the callback for the `param_changed` event.
-    fn param_changed<F>(mut self, callback: F) -> Self
+    pub fn param_changed<F>(mut self, callback: F) -> Self
     where
-        F: Fn(u32, &mut D, *const spa_sys::spa_pod) + 'static,
+        F: FnMut(&StreamRef, &mut D, u32, Option<&spa::pod::Pod>) + 'static,
     {
-        self.callbacks().param_changed = Some(Box::new(callback));
+        self.callbacks.param_changed = Some(Box::new(callback));
         self
     }
 
     /// Set the callback for the `add_buffer` event.
-    fn add_buffer<F>(mut self, callback: F) -> Self
+    pub fn add_buffer<F>(mut self, callback: F) -> Self
     where
-        F: Fn(*mut pw_sys::pw_buffer) + 'static,
+        F: FnMut(&StreamRef, &mut D, *mut pw_sys::pw_buffer) + 'static,
     {
-        self.callbacks().add_buffer = Some(Box::new(callback));
+        self.callbacks.add_buffer = Some(Box::new(callback));
         self
     }
 
     /// Set the callback for the `remove_buffer` event.
-    fn remove_buffer<F>(mut self, callback: F) -> Self
+    pub fn remove_buffer<F>(mut self, callback: F) -> Self
     where
-        F: Fn(*mut pw_sys::pw_buffer) + 'static,
+        F: FnMut(&StreamRef, &mut D, *mut pw_sys::pw_buffer) + 'static,
     {
-        self.callbacks().remove_buffer = Some(Box::new(callback));
+        self.callbacks.remove_buffer = Some(Box::new(callback));
         self
     }
 
     /// Set the callback for the `process` event.
-    fn process<F>(mut self, callback: F) -> Self
+    pub fn process<F>(mut self, callback: F) -> Self
     where
-        F: Fn(&Stream<D>, &mut D) + 'static,
+        F: FnMut(&StreamRef, &mut D) + 'static,
     {
-        self.callbacks().process = Some(Box::new(callback));
+        self.callbacks.process = Some(Box::new(callback));
         self
     }
 
     /// Set the callback for the `drained` event.
-    fn drained<F>(mut self, callback: F) -> Self
+    pub fn drained<F>(mut self, callback: F) -> Self
     where
-        F: Fn() + 'static,
+        F: FnMut(&StreamRef, &mut D) + 'static,
     {
-        self.callbacks().drained = Some(Box::new(callback));
+        self.callbacks.drained = Some(Box::new(callback));
         self
     }
-}
 
-pub struct ListenerLocalBuilder<'a, D> {
-    stream: &'a mut Stream<D>,
-    callbacks: ListenerLocalCallbacks<D>,
-}
-
-impl<'a, D: Default> ListenerBuilderT<D> for ListenerLocalBuilder<'a, D> {
-    fn callbacks(&mut self) -> &mut ListenerLocalCallbacks<D> {
-        &mut self.callbacks
-    }
-}
-
-impl<'a, D> ListenerLocalBuilder<'a, D> {
     //// Register the Callbacks
     ///
     /// Stop building the listener and register it on the stream. Returns a
@@ -640,7 +645,7 @@ impl<'a, D> ListenerLocalBuilder<'a, D> {
             let raw_listener = Box::into_raw(listener);
             let raw_data = Box::into_raw(data);
             pw_sys::pw_stream_add_listener(
-                self.stream.as_ptr(),
+                self.stream.as_raw_ptr(),
                 raw_listener,
                 events.as_ref().get_ref(),
                 raw_data as *mut _,
@@ -655,47 +660,6 @@ impl<'a, D> ListenerLocalBuilder<'a, D> {
     }
 }
 
-pub struct SimpleLocalBuilder<'a, D> {
-    main_loop: &'a MainLoop,
-    name: CString,
-    properties: Properties,
-    callbacks: ListenerLocalCallbacks<D>,
-}
-
-impl<'a, D> ListenerBuilderT<D> for SimpleLocalBuilder<'a, D> {
-    fn callbacks(&mut self) -> &mut ListenerLocalCallbacks<D> {
-        &mut self.callbacks
-    }
-}
-
-impl<'a, D> SimpleLocalBuilder<'a, D> {
-    pub fn create(self) -> Result<Stream<D>, Error> {
-        let (events, data) = self.callbacks.into_raw();
-        let data = Box::into_raw(data);
-        let (stream, mut data) = unsafe {
-            let stream = pw_sys::pw_stream_new_simple(
-                self.main_loop.as_ptr(),
-                self.name.as_ptr(),
-                self.properties.into_raw(),
-                events.as_ref().get_ref(),
-                data as *mut _,
-            );
-            (stream, Box::from_raw(data))
-        };
-        let stream = ptr::NonNull::new(stream).ok_or(Error::CreationFailed)?;
-        data.stream = Some(stream);
-
-        // pw_stream does not keep a pointer on the loop so no need to ensure it stays alive
-        Ok(Stream {
-            ptr: stream,
-            _alive: KeepAlive::Simple {
-                _events: events,
-                _data: data,
-            },
-        })
-    }
-}
-
 pub struct StreamListener<D> {
     listener: Box<spa_sys::spa_hook>,
     // Need to stay allocated while the listener is registered
@@ -706,7 +670,7 @@ pub struct StreamListener<D> {
 impl<D> StreamListener<D> {
     /// Stop the listener from receiving any events
     ///
-    /// Removes the listener registration and cleans up allocated ressources.
+    /// Removes the listener registration and cleans up allocated resources.
     pub fn unregister(self) {
         // do nothing, drop will clean up.
     }
@@ -714,12 +678,13 @@ impl<D> StreamListener<D> {
 
 impl<D> std::ops::Drop for StreamListener<D> {
     fn drop(&mut self) {
-        spa::hook::remove(*self.listener);
+        spa::utils::hook::remove(*self.listener);
     }
 }
 
 bitflags! {
     /// Extra flags that can be used in [`Stream::connect()`]
+    #[derive(Debug, PartialEq, Eq, Clone, Copy)]
     pub struct StreamFlags: pw_sys::pw_stream_flags {
         const AUTOCONNECT = pw_sys::pw_stream_flags_PW_STREAM_FLAG_AUTOCONNECT;
         const INACTIVE = pw_sys::pw_stream_flags_PW_STREAM_FLAG_INACTIVE;
@@ -730,5 +695,7 @@ bitflags! {
         const EXCLUSIVE = pw_sys::pw_stream_flags_PW_STREAM_FLAG_EXCLUSIVE;
         const DONT_RECONNECT = pw_sys::pw_stream_flags_PW_STREAM_FLAG_DONT_RECONNECT;
         const ALLOC_BUFFERS = pw_sys::pw_stream_flags_PW_STREAM_FLAG_ALLOC_BUFFERS;
+        #[cfg(feature = "v0_3_41")]
+        const TRIGGER = pw_sys::pw_stream_flags_PW_STREAM_FLAG_TRIGGER;
     }
 }
